#! /u/feeley/icfp2002/g4/bin/gsi

; file: "server.scm"

(declare
 (standard-bindings)
 (extended-bindings)
 (fixnum)
 (not safe)
 (block)
)

; Inclusion du visualisateur graphique de Marc
(##include "visu.scm")

(define main
  (lambda args
    (if (not (= (length args) 7))
	(main-usage)
	(let* ((port-num (string->number (list-ref args 0)))
	       (nb-players (string->number (list-ref args 1)))
	       (world-file (list-ref args 2))
	       (init-money (string->number (list-ref args 3)))
	       (capacity (string->number (list-ref args 4)))
	       (nb-packs (string->number (list-ref args 5)))
	       (trace? (with-input-from-string (list-ref args 6)
			 (lambda () (read)))))
	  (if (or (not port-num)
		  (not (integer? port-num))
		  (not (exact? port-num))
		  (not nb-players)
		  (not (integer? nb-players))
		  (not (exact? nb-players))
		  (not init-money)
		  (not (integer? init-money))
		  (not (exact? init-money))
		  (not capacity)
		  (not (integer? capacity))
		  (not (exact? capacity))
		  (not nb-packs)
		  (not (integer? nb-packs))
		  (not (exact? nb-packs))
		  (not (boolean? trace?)))
	      (main-usage)
	      (start-server port-num
			    nb-players
			    world-file
			    init-money
			    capacity
			    nb-packs
			    trace?))))))

(define main-usage
  (lambda ()
    (display
     (string-append
      "usage: server.scm <port-num> <nb-players> "
      "<world-file> <money> <capacity> <packages> <trace?>\n"))
    (exit 70)))

(define start-server
  (lambda (port-num nb-players world-file init-money
	   capacity nb-packs trace?)

    (begin
      (display (list ";==> starting a server on port " port-num "\n"))
      (display (list ";==> expecting " nb-players " players\n"))
      (display (list ";==> using world file "))
      (write world-file)
      (display "\n")
      (display
       (list ";==> with " init-money "$ and a capacity of " capacity "\n"))
      (display (list ";==> with " nb-packs " packages and "))
      (display (if trace? "a trace\n" "no trace\n")))

    (let ((listen-port (open-tcp-server port-num '(reuse-address: #t))))
      (let loop ((players '()) (n 0))
	(if (< n nb-players)
	    (let ((player (read listen-port)))
	      (loop (cons player players) (+ n 1)))
	    (begin
	      (close-port listen-port)
	      (start-serving players world-file init-money capacity
			     nb-packs trace?)))))))

(define start-serving
  (lambda (players world-file init-money capacity nb-packs trace?)
    (let ((wport (open-input-file world-file)))
      (let loop ((acc '()))
	(let ((s (read-line wport)))
	  (if (eof-object? s)
	      (begin
		(close-input-port wport)
		(let* ((mapv (list->vector acc))
		       (links (map (lambda (p) (list p p)) players)))
		  (write (server-core mapv init-money capacity
				      nb-packs links trace?))
		  (newline)))
	      (loop (cons s acc))))))))

; ************************************************************
; *** Ancien fichier `serv-in.scm' inclus ici
; ************************************************************

; Fonction principale d'execution de parties
; Arguments:
; - carte: vecteur de string (de la rangee du sud a la rangee du nord)
; - budget: integer
; - capacite: integer
; - nombre de paquets: integer
; - liens vers les clients: ((<in-port> <out-port>) ... (<in-port> <out-port>))
; - trace?: boolean (a savoir si une trace doit etre affichee)
; Retour:
; - scores: (<int> ... <int>)

(define server-core
  (lambda (mapv init-money cap nb-packs links trace?)
    (let* ()
      (set! s-map-v mapv)
      (set! s-map-ref
	    (lambda (x y)
	      (string-ref (vector-ref mapv (- y 1)) (- x 1))))
      (set! s-width (string-length (vector-ref mapv 0)))
      (set! s-height (vector-length mapv))
      (set! s-home-bases (s-extract-home-bases))
      (set! s-cap cap)
      (set! s-nb-robots (length links))
      (set! s-links (list->vector links))
      (set! s-poss (s-choose-init-pos))
      (set! s-money (s-init-money init-money))
      (set! s-nb-packs nb-packs)
      (set! s-packs-state (s-choose-init-packs))
      (set! s-packs-desc (s-choose-packs-desc))
      (set! s-scores (s-init-scores))
      (set! s-load (s-init-load))
      (set! s-booting? (s-init-booting?))
      (set! s-history (s-init-history))
      (set! s-robot-ids (s-first-listen))
      (s-send-board)
      (s-send-confs)
      (s-send-init-resp)
      (if trace?
	  (begin
	    (s-draw-trace-init)
	    (s-draw-trace)))
      (let loop ()
	(if (s-game-finished?)
	    (s-extract-final-scores)
	    (begin
	      (s-reset-booting?!)
	      (s-reset-history)
	      (s-send-local-packs)
	      (let* ((issues (s-collect-issues))
		     (... (s-kill-over-bidders issues))
		     (issues (s-sort-issues issues)))
		(s-execute-issues issues)
		(s-send-turn-resp)
		(if trace?
		    (begin
		      (s-draw-trace-sep)
		      (s-draw-trace)))
		(loop))))))))

; ------------------------------------------------------------

; Variables globales

; Carte brute
; Set by `server'
(define s-map-v #f)

; Nombre de colonnes de la carte
; Set by `server'
(define s-width #f)

; Nombre de rangees de la carte
; Set by `server'
(define s-height #f)

; Vecteur des coordonnnes --- (<int> <int>) --- des "home bases"
; Set by `server'
(define s-home-bases #f)

; Capacite des robots
; Set by `server'
(define s-cap #f)

; Nombre de clients
; Set by `server'
(define s-nb-robots #f)

; Vecteur des liens vers les clients
; Set by `server'
(define s-links #f)

; Vecteur des positions des robots
; position :: (<intx> <inty>) ou dead
(define s-poss #f)

; Nombre de packages
(define s-nb-packs #f)

; Vecteur des descriptions de packages
; description :: (<intx> <inty> <intw>)
(define s-packs-desc #f)

; Vecteur de l'etat des packages
; etat :: #f ou (ground <intx> <inty>) ou (up <intr>)
(define s-packs-state #f)

; Vecteur de la charge de chaque robot
(define s-load #f)

; Vecteur des scores
(define s-scores #f)

; Vecteur de l'argent
(define s-money #f)

; Vecteur de booleens
(define s-booting? #f)

; Vecteur des historiques
; <history> :: (<event> ...) ou #f
; <event> :: <dir>
;            (P <p-id>)
;            (D <p-id>)
(define s-history #f)

; Vecteur des IDs des robots
(define s-robot-ids #f)

; ------------------------------------------------------------

; Fonctions de base

; Acces facile a la carte
; Set by `server'
(define s-map-ref #f)

(define foldr
  (lambda (op base l)
    (if (null? l)
	base
	(op (car l) (foldr op base (cdr l))))))

(define foldr1
  (lambda (op l)
    (let loop ((x (car l)) (l (cdr l)))
      (if (null? l)
	  x
	  (op x (loop (car l) (cdr l)))))))

; Generateur random facile
; Retourne un nombre i t.q. j <= i <= k
(define s-random
  (lambda (j k)
    (+ j (s-rand (- (+ k 1) j)))))

'(define s-make-random-gen
  (lambda (seed)
    (lambda (upto)
      (let* ((hi (quotient seed 127773))
	     (lo (modulo seed 127773))
	     (test (- (* 16807 lo) (* 2836 hi))))
	(if (< test 0)
	    (set! seed (- test -268435456))
	    (set! seed test))
	(modulo seed upto)))))

'(define s-rand (s-make-random-gen 1))

(##include "rand.scm")
(define s-rand rand)

; ------------------------------------------------------------

; Fonctions d'initialisation

; Trouve les "home bases" dans la carte
(define s-extract-home-bases
  (lambda ()
    (let loop ((y 1) (x s-width) (acc '()))
      (cond ((> y s-height)
	     (list->vector acc))
	    ((<= x 0)
	     (loop (+ y 1) s-width acc))
	    ((equal? (s-map-ref x y) #\@)
	     (loop y (- x 1) (cons (list x y) acc)))
	    (else
	     (loop y (- x 1) acc))))))

; Choisit au hasard la position de chaque robot
(define s-choose-init-pos
  (lambda ()
    (let loop ((r-no 1) (acc '()))
      (if (> r-no s-nb-robots)
	  (list->vector acc)
	  (let* ((x (s-random 1 s-width))
		 (y (s-random 1 s-height)))
	    (case (s-map-ref x y)
	      ((#\~ #\# #\@)
	       (loop r-no acc))
	      (else
	       (let ((cpl (list x y)))
		 (if (member cpl acc)
		     (loop r-no acc)
		     (loop (+ r-no 1) (cons cpl acc)))))))))))

; Choisit le positionnement initial des packages (depot)
(define s-choose-init-packs
  (lambda ()
    (let ((nb-homes (vector-length s-home-bases)))
      (let loop ((p-no s-nb-packs) (acc '()))
	(if (<= p-no 0)
	    (list->vector acc)
	    (let* ((h-no (s-random 1 nb-homes))
		   (h-coord (vector-ref s-home-bases (- h-no 1)))
		   (etat (cons 'ground h-coord)))
	      (loop (- p-no 1) (cons etat acc))))))))

; Choisit les caracteristiques des packages
; Note: - verifier la validite de la position de destination
;         (eau, mur, pos. differente)
;       - poids entre 1 et s-cap
(define s-choose-packs-desc
  (lambda ()
    (let loop ((p-no s-nb-packs) (acc '()))
      (if (<= p-no 0)
	  (list->vector acc)
	  (let* ((x (s-random 1 s-width))
		 (y (s-random 1 s-height)))
	    (case (s-map-ref x y)
	      ((#\~ #\#)
	       (loop p-no acc))
	      (else
	       (let ((etat (s-get-pack-state p-no)))
		 (if (and (equal? x (cadr etat)) (equal? y (caddr etat)))
		     (loop p-no acc)
		     (let* ((w (s-random 1 s-cap))
			    (desc (list x y w)))
		       (loop (- p-no 1) (cons desc acc))))))))))))

; ------------------------------------------------------------

; Fonctions communications

; Recoit "Player" de chaque client
; Ne retournait rien
; Retourne #(robot-1-id robot-2-id ... robot-n-id)
(define s-first-listen
  (lambda ()
    (list->vector
     (let loop ((r-no 1))
       (if (<= r-no s-nb-robots)
	   (let* ((line (s-parse-line r-no))
		  (rid (if (null? (cdr line))
			   #f
			   (cadr line))))
	     (cons rid (loop (+ r-no 1))))
	   '())))))

; Envoie le board et ses dimensions a tous
(define s-send-board
  (lambda ()
    (let loop1 ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (begin
	    (s-print-list r-no (list s-width s-height))
	    (let loop2 ((y 1))
	      (if (<= y s-height)
		  (begin
		    (s-print-line r-no (vector-ref s-map-v (- y 1)))
		    (loop2 (+ y 1)))
		  (loop1 (+ r-no 1)))))))))

; Envoie la configuration a chaque client
(define s-send-confs
  (lambda ()
    (let loop ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (let* ((m (s-get-money r-no))
		 (conf (list r-no s-cap m)))
	    (s-print-list r-no conf)
	    (loop (+ r-no 1)))))))

; Envoie la situation initiale a tous
(define s-send-init-resp
  (lambda ()
    (let loop1 ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (let loop2 ((no s-nb-robots) (l '()))
	    (if (<= no 0)
		(begin
		  (s-print-list r-no l)
		  (loop1 (+ r-no 1)))
		(let* ((cpl (s-get-pos no))
		       (x (car cpl))
		       (y (cadr cpl))
		       (s1 (number->string no))
		       (s2 (string-append "#" s1))
		       (l2 (list s2 "X" x "Y" y)))
		  (loop2 (- no 1) (append l2 l)))))))))

; Indique a chaque robot les packages qui sont a ses pieds (s'il n'est
; pas mort)
(define s-send-local-packs
  (lambda ()
    (let loop1 ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (let ((cpl (s-get-pos r-no)))
	    (if (not (equal? cpl 'dead))
		(let* ((x (car cpl))
		       (y (cadr cpl)))
		  (let loop2 ((ps (s-get-ground-packs x y)) (l '()))
		    (if (null? ps)
			(begin
			  (s-print-list r-no l)
			  (loop1 (+ r-no 1)))
			(let* ((p-no (car ps))
			       (desc (s-get-pack-desc p-no)))
			  (loop2 (cdr ps) `(,p-no ,@desc ,@l))))))
		(loop1 (+ r-no 1))))))))

; Collecte les bids et les commandes de chaque client (s'il n'est pas
; mort)
(define s-collect-issues
  (lambda ()
    (let loop ((r-no s-nb-robots) (acc '()))
      (if (<= r-no 0)
	  acc
	  (if (equal? (s-get-pos r-no) 'dead)
	      (loop (- r-no 1) acc)
	      (let* ((l (s-parse-line r-no))
		     (bid (car l))
		     (cmd (cdr l))
		     (issue (list r-no bid cmd)))
		(loop (- r-no 1) (cons issue acc))))))))

; Envoie l'historique de chaque robot qui en a une
(define s-send-turn-resp
  (lambda ()
    (let ((resp-l (s-gather-turn-resp)))
      (let loop1 ((r-no 1))
	(if (<= r-no s-nb-robots)
	    (begin
	      (if (s-get-history r-no)
		  (s-print-list r-no resp-l))
	      (loop1 (+ r-no 1))))))))

(define s-gather-turn-resp
  (lambda ()
    (let loop1 ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (let ((his (s-get-history r-no)))
	    (if his
		(cons
		 (string-append "#" (number->string r-no))
		 (let loop2 ((his his))
		   (if (null? his)
		       (loop1 (+ r-no 1))
		       (let ((ev (car his)))
			 (if (list? ev)
			     (append ev (loop2 (cdr his)))
			     (cons ev (loop2 (cdr his))))))))
		(loop1 (+ r-no 1))))
	  '()))))

; Affiche les scores a la fin de la partie.  Tient compte des IDs des
; robots s'il en ont
(define s-extract-final-scores
  (lambda ()
    (if (s-get-robot-id 1)  ; ont des IDs?
	(let* ((ids (vector->list s-robot-ids))
	       (scores (vector->list s-scores))
	       (alist (map list ids scores))
	       (ass<= (lambda (ass1 ass2) (<= (car ass1) (car ass2))))
	       (sorted-alist (s-sort alist ass<=)))
	  (map cadr sorted-alist))
	(vector->list s-scores))))

; Lit une ligne d'un des liens.  Les donnees sont lues avec `read' et
; mises dans une liste
(define s-parse-line
  (lambda (r-no)
    (read-all (open-input-string (s-read-line r-no)))))

; Lit une ligne venant du port d'entree d'un des robot
(define s-read-line
  (lambda (r-no)
    (read-line (s-get-in-port r-no))))

; Imprime la chaine sur un des liens
(define s-print-line
  (lambda (r-no s)
    (let ((port (s-get-out-port r-no)))
      (display s port)
      (newline port)
      (force-output port))))

; Imprime joliment la liste sur un des liens
(define s-print-list
  (lambda (r-no l)
    (if (null? l)
	(s-print-line r-no "")
	(let* ((ss1 (map s-stringify l))
	       (op (lambda (s l) (cons " " (cons s l))))
	       (ss2 (cons (car ss1) (foldr op '() (cdr ss1))))
	       (s (apply string-append ss2)))
	  (s-print-line r-no s)))))

; Convertit des donnees simples en string (equiv. de `show' en Haskell)
(define s-stringify
  (lambda (obj)
    (cond ((number? obj)
	   (number->string obj))
	  ((string? obj)
	   obj)
	  ((char? obj)
	   (string obj))
	  ((symbol? obj)
	   (symbol->string obj))
	  (else
	   (error "Unknown object type in s-stringify.")))))

; ------------------------------------------------------------

; Manipulation des donnees globales

(define s-get-in-port
  (lambda (r-no)
    (car (s-get-link r-no))))

(define s-get-out-port
  (lambda (r-no)
    (cadr (s-get-link r-no))))

(define s-get-link
  (lambda (r-no)
    (vector-ref s-links (- r-no 1))))

(define s-get-pos
  (lambda (r-no)
    (vector-ref s-poss (- r-no 1))))

(define s-set-pos!
  (lambda (r-no pos)
    (vector-set! s-poss (- r-no 1) pos)))

(define s-get-pack-desc
  (lambda (p-no)
    (vector-ref s-packs-desc (- p-no 1))))

(define s-get-pack-state
  (lambda (p-no)
    (vector-ref s-packs-state (- p-no 1))))

(define s-set-pack-state!
  (lambda (p-no state)
    (vector-set! s-packs-state (- p-no 1) state)))

; Retourne une liste de tous les packages qui sont sur le sol en (x,y)
(define s-get-ground-packs
  (lambda (x y)
    (let loop ((p-no s-nb-packs) (acc '()))
      (if (<= p-no 0)
	  acc
	  (let ((state (s-get-pack-state p-no)))
	    (if state
		(if (equal? (car state) 'ground)
		    (if (and (equal? x (cadr state))
			     (equal? y (caddr state)))
			(loop (- p-no 1) (cons p-no acc))
			(loop (- p-no 1) acc))
		    (loop (- p-no 1) acc))
		(loop (- p-no 1) acc)))))))

(define s-get-money
  (lambda (r-no)
    (vector-ref s-money (- r-no 1))))

(define s-set-money!
  (lambda (r-no money)
    (vector-set! s-money (- r-no 1) money)))

; Met tous les scores a 0
(define s-init-scores
  (lambda ()
    (make-vector s-nb-robots 0)))

(define s-add-score!
  (lambda (r-no w)
    (let ((idx (- r-no 1)))
      (vector-set! s-scores idx (+ (vector-ref s-scores idx) w)))))

; Met tous les loads a 0
(define s-init-load
  (lambda ()
    (make-vector s-nb-robots 0)))

(define s-get-load
  (lambda (r-no)
    (vector-ref s-load (- r-no 1))))

(define s-set-load!
  (lambda (r-no w)
    (vector-set! s-load (- r-no 1) w)))

; Initialise l'argent de chaque robot
(define s-init-money
  (lambda (money)
    (make-vector s-nb-robots money)))

; Valeur initiale du vecteur booting?
(define s-init-booting?
  (lambda ()
    (make-vector s-nb-robots #f)))

; Reinitialise le vecteur de booleens "booting?"
(define s-reset-booting?!
  (lambda ()
    (let loop ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (begin
	    (s-set-booting?! r-no #f)
	    (loop (+ r-no 1)))))))

(define s-get-booting?
  (lambda (r-no)
    (vector-ref s-booting? (- r-no 1))))

(define s-set-booting?!
  (lambda (r-no v)
    (vector-set! s-booting? (- r-no 1) v)))

(define s-ensure-booting?!
  (lambda (r-no)
    (s-set-booting?! r-no #t)))

(define s-init-history
  (lambda ()
    (make-vector s-nb-robots '())))

; Reinitialise le vecteur de l'historique de tous les robots
; Chaque robot vivant a une entree vide, les autres n'en ont pas (#f)
(define s-reset-history
  (lambda ()
    (let loop ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (begin
	    (if (equal? (s-get-pos r-no) 'dead)
		(s-set-history! r-no #f)
		(s-set-history! r-no '()))
	    (loop (+ r-no 1)))))))

(define s-get-history
  (lambda (r-no)
    (vector-ref s-history (- r-no 1))))

(define s-set-history!
  (lambda (r-no v)
    (vector-set! s-history (- r-no 1) v)))

(define s-add-event!
  (lambda (r-no ev)
    (s-set-history! r-no `(,@(s-get-history r-no) ,ev))))

(define s-get-robot-id
  (lambda (r-no)
    (vector-ref s-robot-ids (- r-no 1))))

; ------------------------------------------------------------

; Fonctions de tracage

; Initialisations quelconques
(define s-draw-trace-init
  (lambda ()
    (visu:init! "server" (+ s-width 2) (+ s-height 2))))

; Dessine sur stdout la carte
(define s-draw-trace
  (lambda ()
    (visu:draw-all!)))

; Dessine sur stdout un separateur entre deux traces
(define s-draw-trace-sep
  (lambda ()
    #t))

(define (visu:water? col row)
  (if (or (< col 1) (> col s-width) (< row 1) (> row s-height))
      #f
      (char=? (s-map-ref col row) #\~)))

(define (visu:wall? col row)
  (if (or (< col 1) (> col s-width) (< row 1) (> row s-height))
      #t
      (char=? (s-map-ref col row) #\#)))

(define (visu:base? col row)
  (if (or (< col 1) (> col s-width) (< row 1) (> row s-height))
      #f
      (char=? (s-map-ref col row) #\@)))

(define (visu:robot-id col row)
  (if (or (< col 1) (> col s-width) (< row 1) (> row s-height))
      #f
      (let loop ((r-no 1))
	(if (> r-no s-nb-robots)
	    #f
	    (let ((pos (s-get-pos r-no)))
	      (if (equal? pos 'dead)
		  (loop (+ r-no 1))
		  (if (and (= (car pos) col) (= (cadr pos) row))
		      r-no
		      (loop (+ r-no 1)))))))))

(define (visu:robot-self? id)
  #f)

(define (visu:package-likelihood col row)
  (if (or (< col 1) (> col s-width) (< row 1) (> row s-height))
      0.0
      (let loop ((p-no 1))
	(if (> p-no s-nb-packs)
	    0.0
	    (let ((state (s-get-pack-state p-no)))
	      (if (not state)
		  (loop (+ p-no 1))
		  (if (equal? (car state) 'up)
		      (loop (+ p-no 1))
		      (if (and (= (cadr state) col)
			       (= (caddr state) row))
			  1.0
			  (loop (+ p-no 1))))))))))

(define (visu:robot-package-likelihood r-no)
  (if (> (s-get-load r-no) 0)
      1.0
      0.0))

(define (visu:message)
  (with-output-to-string
    (lambda ()
      (pp 'salut))))

(define (visu:info col row)
  #f)

(define (visu:dot? col row)
  #f)

; ------------------------------------------------------------

; Fonction de tri generique

; Trie les donnees en ordre croissant selon la relation d'ordre donnee.
; Liste de donnees -> liste de donnees
(define s-sort
  (lambda (l rel<=)
    (let ((filter s-sort-filter))
      (let loop ((l l) (acc '()))
	(if (null? l)
	    acc
	    (let* ((pivot (car l))
		   (rst (cdr l))
		   (lows (filter (lambda (d) (rel<= d pivot)) rst))
		   (highs (filter (lambda (d) (not (rel<= d pivot))) rst)))
	      (loop lows (cons pivot (loop highs acc)))))))))

(define s-sort-filter
  (lambda (pred? l)
    (let loop ((l l))
      (if (null? l)
	  '()
	  (let ((d (car l)))
	    (if (pred? d)
		(cons d (loop (cdr l)))
		(loop (cdr l))))))))


; ------------------------------------------------------------

; Fonction de tri non-deterministe

; Trie les issues.  Inclut le non-determinisme pour les egalites
(define s-sort-issues
  (lambda (issues)
    (if (null? issues)
	'()
	(let* ((pivot (car issues))
	       (res (s-partition-issues pivot (cdr issues)))
	       (lows (s-sort-issues (car res)))
	       (highs (s-sort-issues (cadr res))))
	  (append highs (cons pivot lows))))))

(define s-partition-issues
  (lambda (pivot issues)
    (let loop ((issues issues) (lows '()) (highs '()))
      (if (null? issues)
	  (list lows highs)
	  (let ((iss (car issues)))
	    (if (s-<=-issues iss pivot)
		(loop (cdr issues) (cons iss lows) highs)
		(loop (cdr issues) lows (cons iss highs))))))))

(define s-<=-issues
  (lambda (iss1 iss2)
    (let ((bid1 (abs (cadr iss1)))
	  (bid2 (abs (cadr iss2))))
      (if (< bid1 bid2)
	  #t
	  (if (> bid1 bid2)
	      #f
	      (= (s-random 0 1) 0))))))

; ------------------------------------------------------------

; Fonctions secondaires

; Verifie si la partie est terminee
(define s-game-finished?
  (lambda ()
    (or (s-game-finished-by-delivery?)
	(s-game-finished-by-death?))))

(define s-game-finished-by-delivery?
  (lambda ()
    (let loop ((p-no 1))
      (if (<= p-no s-nb-packs)
	  (and (not (s-get-pack-state p-no)) (loop (+ p-no 1)))
	  #t))))

(define s-game-finished-by-death?
  (lambda ()
    (let loop ((r-no 1))
      (if (<= r-no s-nb-robots)
	  (and (equal? (s-get-pos r-no) 'dead) (loop (+ r-no 1)))
	  #t))))

; Tue les robots qui depassent leur budget
; <issue> :: (r-no bid <cmd>)
; <cmd> :: (Move <dir>)
;          (Pick <p-id> ...)
;          (Drop <p-id> ...)
; <dir> :: N | S | E | W
(define s-kill-over-bidders
  (lambda (issues)
    (let loop ((issues issues))
      (if (pair? issues)
	  (let* ((iss (car issues))
		 (r-no (car iss))
		 (bid (cadr iss))
		 (money (s-get-money r-no)))
	    (if (> (abs bid) money)
		(s-execute-death r-no))
	    (loop (cdr issues)))))))

; Execute les issues une a une
(define s-execute-issues
  (lambda (issues)
    (if (pair? issues)
	(let* ((iss (car issues))
	       (r-no (car iss))
	       (bid (cadr iss))
	       (cmd (caddr iss))
	       (pos (s-get-pos r-no)))
	  (if (not (equal? pos 'dead))
	      (begin
		(s-execute-charge r-no (abs bid))
		(if (not (s-get-booting? r-no))
		    (case (car cmd)
		      ((Move)
		       (s-execute-move r-no (cadr cmd)))
		      ((Pick)
		       (s-execute-pick r-no (cdr cmd)))
		      ((Drop)
		       (s-execute-drop r-no (cdr cmd)))
		      (else
		       (error "Unknown command in s-execute-issues."))))))
	  (s-execute-issues (cdr issues))))))

; Fait toute les operations relatives a la mort d'un robot
(define s-execute-death
  (lambda (r-no)
    (s-set-pos! r-no 'dead)
    (let loop ((p-no 1))
      (if (<= p-no s-nb-packs)
	  (let ((state (s-get-pack-state p-no)))
	    (if (and state
		     (equal? (car state) 'up)
		     (= (cadr state) r-no))
		(s-set-pack-state! r-no #f))
	    (loop (+ p-no 1)))))))

; Retire l'argent "biddee"
(define s-execute-charge
  (lambda (r-no m)
    (let ((money (s-get-money r-no)))
      (s-set-money! r-no (- money m)))))

; Tente de prendre les paquets listes
(define s-execute-pick
  (lambda (r-no ps)
    (let ((pos (s-get-pos r-no)))
      (let loop ((ps ps))
	(if (pair? ps)
	    (let ((p-no (car ps)))
	      (if (and (>= p-no 1) (<= p-no s-nb-packs))
		  (let ((state (s-get-pack-state p-no)))
		    (if (and state
			     (equal? (car state) 'ground)
			     (equal? (cdr state) pos))
			(s-execute-try-pick-one r-no p-no))))
	      (loop (cdr ps))))))))

; Tente de prendre le paquet ici present
(define s-execute-try-pick-one
  (lambda (r-no p-no)
    (let* ((desc (s-get-pack-desc p-no))
	   (p-w (caddr desc))
	   (r-w (s-get-load r-no)))
      (if (<= (+ r-w p-w) s-cap)
	  (s-execute-pick-one r-no p-no)))))

; Prend le paquet ici present
(define s-execute-pick-one
  (lambda (r-no p-no)
    (s-set-pack-state! p-no `(up ,r-no))
    (s-set-load! r-no (+ (s-get-load r-no) (caddr (s-get-pack-desc p-no))))
    (s-add-event! r-no `(P ,p-no))))

; Se debarrasse des paquets listes
(define s-execute-drop
  (lambda (r-no ps)
    (if (pair? ps)
	(let ((p-no (car ps)))
	  (if (and (>= p-no 1) (<= p-no s-nb-packs))
	      (let ((state (s-get-pack-state p-no)))
		(if (and state
			 (equal? (car state) 'up)
			 (= (cadr state) r-no))
		    (s-execute-drop-one r-no p-no))))
	  (s-execute-drop r-no (cdr ps))))))

; Se debarrasse de ce paquet
(define s-execute-drop-one
  (lambda (r-no p-no)
    (let* ((pos (s-get-pos r-no))
	   (desc (s-get-pack-desc p-no))
	   (goal? (and (equal? (car pos) (car desc))
		       (equal? (cadr pos) (cadr desc))))
	   (w (caddr desc)))
      (s-set-pack-state! p-no (if goal? #f `(ground ,@pos)))
      (s-set-load! r-no (- (s-get-load r-no) w))
      (if goal? (s-add-score! r-no w))
      (s-add-event! r-no `(D ,p-no)))))

; Effectuer un possible move avec possibilite de push, drop, drown, ...
(define s-execute-move
  (lambda (r-no dir)
    (let* ((pos (s-get-pos r-no))
	   (x (car pos))
	   (y (cadr pos))
	   (delta (case dir
		    ((N) '(0 1)) ((S) '(0 -1)) ((E) '(1 0)) ((W) '(-1 0))))
	   (dx (car delta))
	   (dy (cadr delta))
	   (res (s-execute-inspect (+ x dx) (+ y dy) dx dy))
	   (move? (car res))
	   (pushees (cadr res)))  ; From here to the end
      (s-execute-push pushees)
      (if move?
	  (s-execute-move-really r-no pushees dx dy)))))

; Verifie si le mouvement pourra vraiment se faire et quels robots
; seront pousses par le fait meme.  Retourne (<bool> (<r-id> ...))
(define s-execute-inspect
  (lambda (x y dx dy)
    (if (or (< x 1) (> x s-width) (< y 1) (> y s-height))
	(list #f '())
	(case (s-map-ref x y)
	  ((#\#)
	   (list #f '()))
	  ((#\~)
	   (list #t '()))
	  (else  ; #\@ ou #\.
	   (let ((pushee (let loop ((r-no 1))  ; Cherche si (x,y) est occupee
			   (if (> r-no s-nb-robots)
			       #f
			       (let ((pos (s-get-pos r-no)))
				 (if (equal? pos 'dead)
				     (loop (+ r-no 1))
				     (if (and (equal? (car pos) x)
					      (equal? (cadr pos) y))
					 r-no
					 (loop (+ r-no 1)))))))))
	     (if pushee
		 (let* ((res (s-execute-inspect (+ x dx) (+ y dy) dx dy))
			(move? (car res))
			(pushees (cadr res)))
		   (list move? (cons pushee pushees)))
		 (list #t '()))))))))

; Se fait pousser: provoque un drop random et un reboot
(define s-execute-push
  (lambda (rs)
    (let loop ((rs rs))
      (if (pair? rs)
	  (let ((r-no (car rs)))
	    (s-execute-drop-random r-no)
	    (s-ensure-booting?! r-no)
	    (loop (cdr rs)))))))

; Si le robot a au moins un paquet, il en perd un au hasard
(define s-execute-drop-random
  (lambda (r-no)
    (let ((ps (let loop ((p-no 1))
		(if (> p-no s-nb-packs)
		    '()
		    (let ((state (s-get-pack-state p-no)))
		      (if (and state
			       (equal? (car state) 'up)
			       (= (cadr state) r-no))
			  (cons p-no (loop (+ p-no 1)))
			  (loop (+ p-no 1))))))))
      (if (pair? ps)
	  (let ((p-no (list-ref ps (- (s-random 1 (length ps)) 1))))
	    (s-execute-drop-one r-no p-no))))))

; Effectue tous les deplacements, du plus lointain au robot "actif"
(define s-execute-move-really
  (lambda (r-no rs dx dy)
    (let loop ((rs (reverse (cons r-no rs))))
      (if (pair? rs)
	  (begin
	    (s-execute-move-one-really (car rs) dx dy)
	    (loop (cdr rs)))))))

; Effectue le deplacement (sans obstacles) d'un robot
(define s-execute-move-one-really
  (lambda (r-no dx dy)
    (let* ((pos (s-get-pos r-no))
	   (x (+ (car pos) dx))
	   (y (+ (cadr pos) dy))
	   (... (s-set-pos! r-no (list x y)))
	   (ev (case dx ((-1) 'W) ((0) (case dy ((-1) 'S) ((1) 'N))) ((1) 'E)))
	   (... (s-add-event! r-no ev)))
      (if (equal? (s-map-ref x y) #\~)
	  (s-execute-death r-no)))))

(apply main (cdr (command-line)))

(force-output (current-output-port)) ; **** IMPORTANT ****
