; file: "visu.scm"

(load "x11.o1")

;! Largeur du monde en nombre de cases
(define visu:width 0)
;! Hauteur du nombre en nombre de cases
(define visu:height 0)

;! Nombre de pixel de haut/large pour chaque case
(define visu:size 0)

;! Nombre de pixel de haut/large pour la ligne de statut
(define visu:status 0)

;! L'etat de l'ecran
(define visu:screen-state #f)

;! Initialise la fenetre de visualisation
;! @param width  la largeur en nombre de cases (un entier)
;! @param height la hauteur en nombre de cases (un entier)
;! @return @c (void)
(define (visu:init! title width height)

  (define size 16)
  (define status 100)

  (set! visu:width width)
  (set! visu:height height)
  (set! visu:size size)
  (set! visu:status status)
  (set! visu:screen-state (make-vector (* width height) 'undefined))

  (g-init title (* width size) (+ status (* height size))))

(define (visu:draw-all!)

  (g-box white
	 0
	 (* visu:height visu:size)
	 (* visu:width visu:size)
         visu:status)

  (let ((lst (read-all (open-input-string (visu:message)) read-line)))
    (let loop ((lst lst) (i 1))
      (if (pair? lst)
          (let ((m (car lst)))
            (g-text black
                    left
                    5
                    (- (+ (* visu:height visu:size) visu:status)
                       (* i 10))
                    m)
            (loop (cdr lst) (+ i 1))))))

  (let loop-row ((row 0))
    (if (< row visu:height)
        (let loop-col ((col 0))
          (if (< col visu:width)
	      (begin
		(visu:draw-cell col row)
                (loop-col (+ col 1)))
              (loop-row (+ row 1))))))
  (g-flush))

(define (visu:draw-cell col row)
  (cond ((visu:water? col row)
         (if (visu:update? col row 'water)
             (visu:draw-box col row blue 0)))
        ((visu:wall? col row)
         (if (visu:update? col row 'wall)
             (visu:draw-box col row black 0)))
        (else
          (let ((id (visu:robot-id col row)))
            (if (or (let ((x (visu:package-likelihood col row)))
                      (cond ((##flonum.= x 0.00)
                             (if (visu:update? col row 0)
                                 (begin
                                   (visu:draw-box col row white 0)
                                   #t)
                                 #f))
                            ((##flonum.< x 0.33)
                             (if (visu:update? col row 1)
                                 (begin
                                   (visu:draw-box col row pink0 0)
                                   #t)
                                 #f))
                            ((##flonum.< x 0.67)
                             (if (visu:update? col row 2)
                                 (begin
                                   (visu:draw-box col row pink1 0)
                                   #t)
                                 #f))
                            ((##flonum.< x 1.00)
                             (if (visu:update? col row 3)
                                 (begin
                                   (visu:draw-box col row pink2 0)
                                   #t)
                                 #f))
                            (else
                             (if (visu:update? col row 4)
                                 (begin
                                   (visu:draw-box col row pink3 0)
                                   #t)
                                 #f))))
                    id)
                (if id
                    (let ((x (visu:robot-package-likelihood id))
                          (self? (visu:robot-self? id)))
                      (visu:update? col row (cons 1 2))
                      (cond ((##flonum.= x 0.00)
                             (visu:draw-box col row green0 1))
                            ((##flonum.< x 0.50)
                             (visu:draw-box col row green1 1))
                            ((##flonum.< x 1.00)
                             (visu:draw-box col row green2 1))
                            (else
                             (visu:draw-box col row green3 1)))
                      (if self?
                          (visu:draw-box col row black (- (quotient visu:size 2) 4)))
                      (g-text (if self? white black)
                              center
                              (+ (* col visu:size) (quotient visu:size 2))
                              (+ (* row visu:size) (quotient visu:size 2))
                              (number->string id)))
                    (let ((info (visu:info col row)))
                      (visu:draw-box col row white 1)
                      (if info
                          (begin
                            (visu:update? col row (cons 1 2))
                            (g-text magenta
                                    center
                                    (+ (* col visu:size) (quotient visu:size 2))
                                    (+ (* row visu:size) (quotient visu:size 2))
                                    info))))))
            (if (visu:base? col row)
                (visu:draw-box col row cyan (- (quotient visu:size 2) 2)))
            (if (visu:dot? col row)
                (begin
                  (visu:update? col row (cons 1 2))
                  (visu:draw-box col row red (- (quotient visu:size 2) 1))))))))

(define (visu:update? col row new-state)
  (let* ((index (+ col (* row visu:width)))
         (state (vector-ref visu:screen-state index)))
    (if (eq? state new-state)
        #f
        (begin
          (vector-set! visu:screen-state index new-state)
          #t))))

(define (visu:draw-box col row color border)
  (g-box color
	 (+ (* col visu:size) border)
	 (+ (* row visu:size) border)
	 (- visu:size (* 2 border))
	 (- visu:size (* 2 border))))

#|

(define (get-cell col row)
  (string-ref (vector-ref visu:map row) col))

(define (visu:water? col row) (char=? (get-cell col row) #\~))
(define (visu:wall? col row)  (char=? (get-cell col row) #\#))
(define (visu:base? col row)  (char=? (get-cell col row) #\@))
(define (visu:package-likelihood col row pack?) (if pack? 1.0 (/ (modulo (+ col row) 10) 10)))

(define (draw-robot col row w id)
  (let ((i-1 (- i 1))
	(j-1 (- j 1)))
    (visu:draw-box i-1 j-1 white 0)
    (visu:draw-box i-1 j-1 11 2)
    (g-text black
	    center
	    (+ (* j-1 visu:size) (/ visu:size 2))
	    (+ (* i-1 visu:size) (/ visu:size 2))
	    (number->string id))))

(define (draw-package col row w id)
  (let ((i-1 (- i 1))
	(j-1 (- j 1)))
    (visu:draw-cell i-1 j-1 #t)))

;! Met la map a jour
;! @param robots nouvelles positions des robots. C'est une liste de robot, ou 
;!   chaque robot est une liste de la forme (id x y poids nb-package montant-restant).
;! @param packages nouvelles positions des packages. C'est une liste de packages, 
;!   ou chaque package est une liste de la forme (id x y poids)
(define (visu:update! robots packages)
  ;; enleve les robots de la map.
  (if visu:last-positions
      (for-each 
       (lambda (robot)
	 (let ((x (- (list-ref robot 1) 1))
	       (y (- (list-ref robot 2) 1)))
	   (visu:draw-cell x y #f)))
       visu:last-positions))

  (if visu:last-packages
      (for-each 
       (lambda (pack)
	 (let ((x (- (list-ref pack 1) 1))
	       (y (- (list-ref pack 2) 1)))
	   (visu:draw-cell x y #f)))
       visu:last-packages))

  (for-each
   (lambda (pack)
     (let ((i (list-ref pack 0))
	   (x (list-ref pack 1))
	   (y (list-ref pack 2))
	   (w (list-ref pack 3)))
       (draw-package x y w i)))
   packages)

  (for-each 
   (lambda (robot)
     (let ((i (list-ref robot 0))
	   (x (list-ref robot 1))
	   (y (list-ref robot 2))
	   (w (list-ref robot 3))
	   (n (list-ref robot 4))
	   (m (list-ref robot 5)))
       (draw-robot x y w i)))
   robots)

  (set! visu:last-positions robots)
  (set! visu:last-packages packages))
      



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; A ENLEVER
(define (test1)
  (visu:init! 21 20 16
   '#("....................."
      ".........#.#........."
      "~~~~~~~~~#.#~~~~~~~~~"
      "~~~~~~~~~#.#~~~~~~~.."
      "~~~~.....#.#...~~~..."
      ".~~......#.#........."
      ".........#.#........."
      "......####.####......"
      "......#.......#......"
      "#######.......#######"
      "..........@.........."
      "#######.......#######"
      "......#.......#......"
      "......####.####......"
      ".........#.#........."
      ".........#.#........."
      ".........#.#...~~...."
      ".........#.#........."
      ".........#.#........."
      ".........#.#........."))
  
  (read-char)
  (for-each
   (lambda (r p)
     (visu:update! r p)
     (read-char))
   '(((10 9 9 35 2 100))
     ((10 10 9 35 2 100))
     ((10 11 9 35 2 100))
     ((10 11 10 35 2 100)))
   '(((1 11 11 4))
     ((1 11 11 4))
     ((1 11 11 4))
     ((1 11 11 4)))))

(define (test3)
  (let ((map 
	 (vector
	  "############################################################################"
	  "#..#..#.....#.....#........#.....#..#.................#..#.................#"
	  "#..#..####..#..#..####..#######..#..#..#######..#######..####..####..#..#..#"
	  "#..#.....#.....#..#.....#..#........#..#..#........#..#..#..#..#.....#..#..#"
	  "#..#..#######..#######..#..####..#######..#######..#..#..#..#############..#"
	  "#..#.....#..#.....#.....#..#......................................#..#.....#"
	  "#..#..####..#..#######..#..################..#..#..#..#######..####..####..#"
	  "#........#..#........#..............#.....#..#..#..#..#....................#"
	  "#..####..#..#######..#..#######..#######..#..#############..#######..####..#"
	  "#..#...........#..#........#..#..............#..............#...........#..#"
	  "#######..#..####..#..#..####..#..#######..####..#..####..##########..#..####"
	  "#........#...........#.....#.....#........#..#..#..#..#..#...........#.....#"
	  "####..#..##########..####..#..##########..#..#..####..##########..#..#..####"
	  "#.....#.....#...........#..#..#.....#........#..#........#.....#..#..#.....#"
	  "#..#..######################..####..################..####..##########..####"
	  "#..#..#..#..#..#..#..............#..#..#..#..#.................#...........#"
	  "#..#..#..#..#..#..##########..####..#..#..#..####..##########..####..#######"
	  "#..#..#..............#.....#..#..#.................#.................#..#..#"
	  "#######..#..#######..#..####..#..#######..#######..################..#..#..#"
	  "#..#..#..#..#...........#.....#..............#........#....................#"
	  "#..#..#..##########..#..#######..####..#..####..#..####..#######..#..####..#"
	  "#.....#..#..#........#..#..#.....#..#..#..#.....#..#.....#........#.....#..#"
	  "####..#..#..##########..#..#######..#..#..##########..#######..#..####..####"
	  "#.....#..#..#.....#....................#..#.................#..#.....#.....#"
	  "####..#..#..####..####..#..#..####..####..#######..#..##########..##########"
	  "#........#..#..#..#..#..#..#..#.....#.....#........#...........#........#..#"
	  "####..####..#..#..#..#..#######..#..####..####..####..##########..####..#..#"
	  "#..#.....#.................#.....#..#.....#..#.....#........#........#.....#"
	  "#..#############..####..##########..#######..####..#..#..#..####..####..####"
	  "#..#..#.....#.....#........#..#..#..#..............#..#..#..#..#.....#..#..#"
	  "#..#..####..####..#..#######..#..##########..#..##########..#..#..####..#..#"
	  "#..#..#...........#..#..#....................#...........#..#........#.....#"
	  "#..#..#..#############..####..####..#..####..####..##########..#######..#..#"
	  "#...........#.....#.....#..#..#.....#..#.....#.....#..#.....#..#..#.....#..#"
	  "####..##########..####..#..#######..#..#..#..####..#..####..#..#..####..####"
	  "#........#..........................#..#..#.....#..#..#..#........#........#"
	  "#..#..##########..#..#..#..####..#######..#..#..####..#..####..#..####..####"
	  "#..#.....#........#..#..#.....#..#........#..#..#..............#..#........#"
	  "#..#..##########..#######..####..#######..####..#..#..#..#..#..#######..#..#"
	  "#..#..#........#..#........#..#..#..#.....#.....#..#..#..#..#.....#.....#..#"
	  "####..#..####..##########..#..#..#..#..####..#..####..#..#..#..####..#..####"
	  "#.....#..#........#..#..#.....#.....#..#..#..#.....#..#..#..#..#.....#..#..#"
	  "####..#######..####..#..#############..#..####..#######..#..#######..#..#..#"
	  "#.....#.....#..........................#..............#..#..#........#.....#"
	  "####..####..####..####..####..####..####..#..#..#######..####..#..#..#..####"
	  "#.....#........#..#.....#..#..#........#..#..#.....#........#..#..#..#.....#"
	  "##########..#..#######..#..#..#######..####..####..##########..#..#..#..#..#"
	  "......#.....#..#........#.....#..#..#..#..#..#........#.....#..#..#..#..#..#"
	  "...#..#######..####..##########..#..####..#..####..#######@@#..####..####@@#"
	  "...#....................#.......................#..#......@@......#..#...@@#"
	  "...#########################################################################")))
    (visu:init! (string-length (vector-ref map 0))
		 (vector-length map)
		 10 
		 map)
    (read-char)))

'(define (test2)
  (visu:update! 
   '#("~~~~~~~~~#.#~~~~~~~~~"
      "~~~~~~~~~#.#~~~~~~~~~"
      "~~~~~~~~~#.#~~~~~~~~~"
      "~~~~~~~~~#.#~~~~~~~~~"
      "~~~~~~~~~#.#~~~~~~~~~"
      "~~~~~~~~~#.#~~~~~~~~~"
      "~~~~~~####.####~~~~~~"
      "~~~~~~#.......#~~~~~~"
      "~#~#~##..@@@..##~#~#~"
      ".........@@@........."
      ".........@@@........."
      "#~#~#~#..@@@..#~#~#~#"
      "~~~~~~#.......#~~~~~~"
      "......####.####......"
      ".........#.#........."
      ".........#.#........."
      ".........#.#........."
      ".........#.#........."
      ".........#.#........."
      ".........#.#........."
      ".....................") '() '()))

(test1)
(test3)


(force-output (current-output-port)) ; **** IMPORTANT ****


|#
